import abc
import asyncio
import inspect
import time
import typing
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Union

import typing_extensions

from . import api_types, types
from .async_utils import run_as_coro
from .datasets import Row
from .types import _EvalParent

EVALUATION_ARGS = [
    "row",
    "task_result",
    "evaluated_model_system_prompt",
    "evaluated_model_retrieved_context",
    "evaluated_model_input",
    "evaluated_model_output",
    "evaluated_model_gold_answer",
    "evaluated_model_attachments",
    "tags",
    "parent",
    "experiment_id",
    "dataset_id",
    "dataset_sample_id",
]

EvalF = typing.TypeVar("EvalF", bound=typing.Callable[..., typing.Any])


class EvaluatorError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class Evaluator(abc.ABC):
    name: str = ""
    criteria: Optional[str] = None
    accepted_args: set[str]
    remote_capture = False

    def __init__(
        self,
        accepted_args: Optional[set[str]] = None,
        *,
        evaluator_name: Optional[str] = None,
        criteria: Optional[str] = None,
    ):
        self.name = self.name or evaluator_name or self.__class__.__name__
        self.criteria = self.criteria or criteria
        self.accepted_args = accepted_args
        if not self.accepted_args:
            sig = inspect.signature(self.evaluate)
            param_keys = sig.parameters.keys()
            for name in param_keys:
                if name not in EVALUATION_ARGS:
                    raise ValueError(
                        f"{name!r} is not a valid evaluator argument. Valid arguments are: {EVALUATION_ARGS}"
                    )
            self.accepted_args = set(param_keys)

    def __repr__(self):
        return f"<Evaluator (id: {self.name!r}, criteria: {self.criteria!r}) of class {self.__class__.__name__}>"

    def display_name(self) -> str:
        if not self.criteria:
            return self.name
        return f"{self.name}:{self.criteria}"

    @property
    def column_prefix(self):
        """
        Returns a column prefix that is used in exported dataframe for result columns created by the evaluator.
        """
        p = self.criteria or ""
        return f"{self.name}.{p}"

    def __call__(self, *args, **kwargs):
        return self.evaluate(*args, **kwargs)

    async def execute(
        self,
        *,
        loop: asyncio.AbstractEventLoop,
        executor: ThreadPoolExecutor,
        row: Row,
        task_result: Optional[types.TaskResult],
        tags: dict[str, str],
        parent: _EvalParent,
        experiment_id: str,
        dataset_id: Optional[str] = None,
        dataset_sample_id: Optional[int] = None,
    ) -> Optional[types.EvaluatorOutput]:
        model_output = task_result and task_result.evaluated_model_output or row.evaluated_model_output
        kwargs = {
            "evaluated_model_system_prompt": row.evaluated_model_system_prompt,
            "evaluated_model_retrieved_context": row.evaluated_model_retrieved_context,
            "evaluated_model_input": row.evaluated_model_input,
            "evaluated_model_output": model_output,
            "evaluated_model_gold_answer": row.evaluated_model_gold_answer,
            "evaluated_model_attachments": row.evaluated_model_attachments,
            "experiment_id": experiment_id,
            "row": row,
            "task_result": task_result,
            "parent": parent,
            "dataset_id": dataset_id,
            "dataset_sample_id": dataset_sample_id,
            "tags": tags,
        }
        pass_kwargs = {k: v for k, v in kwargs.items() if k in self.accepted_args}
        start_time = time.perf_counter()
        result = await run_as_coro(
            __loop=loop,
            __executor=executor,
            __fn=self.evaluate,
            **pass_kwargs,
        )
        if result is None:
            return None

        end_time = time.perf_counter()
        elapsed = end_time - start_time

        if isinstance(result, bool):
            result = types.EvaluationResult(pass_=result, score_raw=float(result))
        if isinstance(result, (int, float)):
            result = types.EvaluationResult(pass_=None, score_raw=float(result))
        if isinstance(result, api_types.EvaluationResult):
            elapsed = result.evaluation_duration and result.evaluation_duration.seconds
        if isinstance(result, types.EvaluationResult):
            if not result.evaluation_duration_s:
                result.evaluation_duration_s = elapsed
        if not isinstance(result, (types.EvaluationResult, api_types.EvaluationResult)):
            raise ValueError(
                f"Evaluator {self.display_name()!r} returned an object of unexpected type {type(result)!r}"
            )

        return types.EvaluatorOutput(result=result, duration=elapsed)

    @abc.abstractmethod
    def evaluate(self, **kwargs): ...


class SyncFunctionalEvaluator(Evaluator):
    fn: EvalF

    def __init__(self, *, evaluator_name: str, criteria: Optional[str] = None, fn: EvalF, accepted_args: set[str]):
        self.fn = fn
        super().__init__(accepted_args, evaluator_name=evaluator_name, criteria=criteria)

    def evaluate(self, **kwargs):
        return self.fn(**kwargs)


class FunctionalEvaluator(Evaluator):
    fn: EvalF

    def __init__(self, *, evaluator_name: str, criteria: Optional[str] = None, fn: EvalF, accepted_args: set[str]):
        self.fn = fn
        super().__init__(accepted_args, evaluator_name=evaluator_name, criteria=criteria)

    async def evaluate(self, **kwargs):
        return await self.fn(**kwargs)


def evaluator(
    func=None, *, name: str = None, criteria: str = None
) -> Union[Evaluator, typing.Callable[[...], Evaluator]]:
    def decorator(fn: EvalF) -> Evaluator:
        sig = inspect.signature(fn)
        param_keys = sig.parameters.keys()
        for param_name in param_keys:
            if param_name not in EVALUATION_ARGS:
                raise ValueError(
                    f"{param_name!r} is not a valid evaluator argument. Valid arguments are: {EVALUATION_ARGS}"
                )
        evaluator_name = name or fn.__name__
        if inspect.iscoroutinefunction(fn):
            return FunctionalEvaluator(
                evaluator_name=evaluator_name,
                criteria=criteria,
                fn=fn,
                accepted_args=set(param_keys),
            )
        else:
            return SyncFunctionalEvaluator(
                evaluator_name=evaluator_name,
                criteria=criteria,
                fn=fn,
                accepted_args=set(param_keys),
            )

    if func is None:
        return decorator
    return decorator(func)


@typing_extensions.deprecated(
    "'simple_evaluator' is deprecated and will be removed in future versions. Use '@evaluator' decorator instead."
)
def simple_evaluator(fn: typing.Callable[[str, str], bool], name: Optional[str] = None) -> FunctionalEvaluator:
    """
    Simple evaluator allows to wrap functions boolean evaluation:

    ```python
    exact_match = simple_evaluator(lambda output, gold_answer: output == gold_answer)
    ```

    It can be also used to decorate multi-line function.

    ```python
    @simple_evaluator
    def iexact_match(output, gold_answer) -> bool:
        output = output.strip().lower()
        gold_answer = gold_answer.strip().lower()
        return output == gold_answer
    ```
    """

    def wrapper(evaluated_model_output: str, evaluated_model_gold_answer: str):
        passed = fn(evaluated_model_output, evaluated_model_gold_answer)
        return types.EvaluationResult(pass_=passed, score_raw=float(passed))

    name = name or fn.__name__
    if name == "<lambda>":
        name = "simple_evaluator"
    return evaluator(wrapper, evaluator_name=name)



import typing
from typing import Optional, Union

import pydantic

from . import api_types
from .api_types import EvaluateResult


class TaskResult(pydantic.BaseModel):
    evaluated_model_output: str
    metadata: Optional[dict[str, typing.Any]] = None
    tags: Optional[dict[str, str]] = None


class EvaluationResult(pydantic.BaseModel):
    pass_: Optional[bool] = pydantic.Field(default=False, serialization_alias="pass")
    score_raw: Optional[float] = None
    text_output: Optional[str] = None
    explanation: Optional[str] = None
    metadata: Optional[dict[str, typing.Any]] = None
    tags: Optional[dict[str, str]] = None
    evaluation_duration_s: Optional[float] = None
    explanation_duration_s: Optional[float] = None


class LocalEvaluator(typing.Protocol):
    def __call__(
        self,
        evaluated_model_system_prompt: Optional[str],
        evaluated_model_retrieved_context: Optional[Union[typing.List[str], str]],
        evaluated_model_input: Optional[str],
        evaluated_model_output: Optional[str],
        evaluated_model_gold_answer: Optional[str],
        evaluated_model_attachments: Optional[list[dict[str, typing.Any]]],
        tags: Optional[typing.Dict[str, str]] = None,
        explain_strategy: typing.Literal["never", "on-fail", "on-success", "always"] = "always",
        **kwargs,
    ) -> EvaluationResult: ...


class EvaluatorOutput(pydantic.BaseModel):
    result: Union[EvaluationResult, api_types.EvaluationResult]
    duration: float


MaybeEvaluationResult = typing.Union[EvaluateResult, api_types.EvaluationResult, None]


class EvalsMap(dict):
    def __contains__(self, item) -> bool:
        item = self._key(item)
        return super().__contains__(item)

    def __getitem__(self, item) -> MaybeEvaluationResult:
        item = self._key(item)
        return super().__getitem__(item)

    def __setitem__(self, key: str, value: MaybeEvaluationResult):
        key = self._key(key)
        return super().__setitem__(key, value)

    @staticmethod
    def _key(item):
        if isinstance(item, str):
            return item
        if hasattr(item, "display_name"):
            return item.display_name()
        return item


class _EvalParent(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    task: Optional[TaskResult]
    evals: typing.Optional[EvalsMap]
    parent: typing.Optional["_EvalParent"]

    def find_eval_result(self, evaluator_or_name) -> Union[api_types.EvaluationResult, EvaluationResult, None]:
        if not self.evals and self.parent:
            return self.parent.find_eval_result(evaluator_or_name)
        if evaluator_or_name in self.evals:
            return self.evals[evaluator_or_name]
        return None


_EvalParent.model_rebuild()

EvalParent = typing.Optional[_EvalParent]



import asyncio
import functools
import random
import threading
from concurrent.futures.thread import ThreadPoolExecutor

finished = 0
exs = 0

async def dummy_2(x):
    global finished
    finished += 1
    return x + 2

async def dummy(x):
    if random.random() < 0.01:
        global exs
        exs += 1
        raise ValueError("??")
    global finished
    finished += 1
    return (await dummy_2(x)) + 1


class Runner:
    def __init__(self):
        self.loop = asyncio.get_event_loop_policy().new_event_loop()
        self.executor = ThreadPoolExecutor()
        self.thread = threading.Thread(
            target=self._run_loop,
            daemon=False,
        )
        self.thread.start()
        print("Running...", flush=True)

    def _run_loop(self):
        self.loop.run_forever()
        print(f"run_forever stopped, finished: {finished}", flush=True)
        tasks = asyncio.all_tasks(self.loop)
        if tasks:
            self.loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))
            print(f"run_until_completed, finished: {finished}", flush=True)
        self.loop.close()
        print(f"excs: {exs}", flush=True)


    def submit(self, fn, *args, **kwargs):
        if asyncio.iscoroutinefunction(fn):
            asyncio.run_coroutine_threadsafe(fn(*args, **kwargs), self.loop)
        else:
            asyncio.run_coroutine_threadsafe(functools.partial(self.loop.run_in_executor, fn, args, kwargs), self.loop)

    def shutdown(self):
        print("Shutting down...", flush=True)
        self.loop.call_soon_threadsafe(self.loop.stop)
        #
        # for _ in range(1000):
        #     print(f"Running: {self.loop.is_running()}", flush=True)
        #     if not self.loop.is_running():
        #         break
        #
        #
        # pending_tasks = [t for t in asyncio.all_tasks(self.loop) if not t.done()]
        # if not pending_tasks:
        #     return
        # print(f"Pending tasks left: {len(pending_tasks)}")
        # self.loop.run_until_complete(asyncio.gather(*pending_tasks))
        #
        # print("Closing the loop")
        # self.loop.close()

def main():
    runner = Runner()

    for x in range(3000):
        runner.submit(dummy, x)

    runner.shutdown()
    print(f"Post shutdown; finished: {finished}", flush=True)
    print("MAIN EXIT", flush=True)


if __name__ == "__main__":
    main()


import asyncio
import dataclasses
import json
import pathlib
import re
import typing
from typing import Optional, Union

import pandas as pd
import typing_extensions as te


class Fields(typing.TypedDict, total=False):
    sid: te.NotRequired[Optional[int]]
    evaluated_model_system_prompt: te.NotRequired[Optional[str]]
    evaluated_model_retrieved_context: te.NotRequired[Union[str, list[str], None]]
    evaluated_model_input: te.NotRequired[Optional[str]]
    evaluated_model_gold_answer: te.NotRequired[Optional[str]]
    evaluated_model_name: te.NotRequired[Optional[str]]
    evaluated_model_provider: te.NotRequired[Optional[str]]
    evaluated_model_selected_model: te.NotRequired[Optional[str]]
    evaluated_model_params: te.NotRequired[Optional[dict[str, typing.Any]]]


@dataclasses.dataclass
class Row:
    _row: pd.Series

    def __getattr__(self, name: str):
        return self._row[name]

    @property
    def row(self) -> pd.Series:
        return self._row

    @property
    def dataset_id(self) -> Optional[str]:
        return self._row.get("dataset_id")

    @property
    def sid(self) -> int:
        return self._row.sid

    @property
    def evaluated_model_system_prompt(self) -> Optional[str]:
        if "evaluated_model_system_prompt" in self._row.index:
            return self._row.evaluated_model_system_prompt
        return None

    @property
    def evaluated_model_retrieved_context(self) -> Optional[list[str]]:
        ctx = None
        if "evaluated_model_retrieved_context" in self._row.index:
            ctx = self._row.evaluated_model_retrieved_context
        if ctx is None:
            return None
        if isinstance(ctx, str):
            return [ctx]
        assert isinstance(ctx, list), f"evaluated_model_retrieved_context is not a list, its: {type(ctx)}"
        return ctx

    @property
    def evaluated_model_input(self) -> Optional[str]:
        if "evaluated_model_input" in self._row.index:
            return self._row.evaluated_model_input
        return None

    @property
    def evaluated_model_output(self) -> Optional[str]:
        if "evaluated_model_output" in self._row.index:
            return self._row.evaluated_model_output
        return None

    @property
    def evaluated_model_gold_answer(self) -> Optional[str]:
        if "evaluated_model_gold_answer" in self._row.index:
            return self._row.evaluated_model_gold_answer
        return None

    @property
    def evaluated_model_attachments(self) -> Optional[list[dict[str, typing.Any]]]:
        if "evaluated_model_attachments" in self._row.index:
            return self._row.evaluated_model_attachments
        return None


@dataclasses.dataclass
class Dataset:
    dataset_id: Optional[str]
    df: pd.DataFrame

    def iterrows(self) -> typing.Iterable[Row]:
        for i, row in self.df.iterrows():
            yield Row(row)

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, dataset_id: Optional[str] = None) -> te.Self:
        df = cls.__sanitize_df(df, dataset_id)
        return cls(df=df, dataset_id=dataset_id)

    @classmethod
    def from_records(
        cls,
        records: Union[typing.Iterable[Fields], typing.Iterable[dict[str, typing.Any]]],
        dataset_id: Optional[str] = None,
    ) -> te.Self:
        df = pd.DataFrame.from_records(records)
        df = cls.__sanitize_df(df, dataset_id)
        return cls(df=df, dataset_id=dataset_id)

    @classmethod
    def __sanitize_df(cls, df: pd.DataFrame, dataset_id: str) -> pd.DataFrame:
        # Validate and backfill "sid"
        if "sid" not in df.columns:
            df["sid"] = range(1, len(df) + 1)

        sid_count = df["sid"].count()
        if sid_count != 0 and sid_count != len(df):
            raise ValueError("Either all entries must have 'sid' or none should have it.")

        if sid_count == 0:
            df["sid"] = range(1, len(df) + 1)
        else:
            if df["sid"].duplicated().any():
                raise ValueError("Duplicate 'sid' values found. Dataset cannot have duplicated 'sid's.")

        if not pd.api.types.is_integer_dtype(df["sid"]):
            try:
                df["sid"] = df["sid"].astype(int)
            except ValueError:
                raise ValueError("'sid' column contains non-integer values that cannot be converted to integers.")

        def normalize_context(value) -> Optional[list[str]]:
            if value is None:
                return None

            if isinstance(value, list):
                return [str(v) for v in value if v]

            if pd.isna(value) or value == "" or value == "nan":
                return None

            if isinstance(value, str):
                try:
                    parsed = json.loads(value)
                    if isinstance(parsed, list):
                        return [str(v) for v in parsed if v]
                    else:
                        return [str(parsed)]
                except json.JSONDecodeError:
                    return [value]

            return [str(value)]

        def _assert_attachment(value: dict):
            assert isinstance(
                value.get("url"), str
            ), "parsing 'evaluated_model_attachments': missing or invalid type (expected str) of 'url' field"
            assert isinstance(
                value.get("media_type"), str
            ), "parsing 'evaluated_model_attachments': missing or invalid type (expected str) of 'media_type' field"
            usage_type = value.get("usage_type")
            assert (
                isinstance(usage_type, str) or usage_type is None
            ), "parsing 'evaluated_model_attachments': invalid type (expected str) of 'usage_type' field"

        def _normalize_attachment(value) -> dict[str, typing.Any]:
            if isinstance(value, dict):
                _assert_attachment(value)
                return value

        def normalize_attachments(value) -> Optional[list[dict[str, typing.Any]]]:
            if value is None:
                return None

            if isinstance(value, list):
                return [_normalize_attachment(v) for v in value]

            if isinstance(value, dict):
                _assert_attachment(value)
                return [value]

            if isinstance(value, str):
                try:
                    return normalize_attachments(json.loads(value))
                except json.JSONDecodeError as exc:
                    raise ValueError(f"parsing 'evaluated_model_attachments': {exc}")

            raise ValueError("parsing 'evaluated_model_attachments': unexpected value type")

        if "evaluated_model_system_prompt" in df.columns:
            df["evaluated_model_system_prompt"] = df["evaluated_model_system_prompt"].astype("string[python]")
            df["evaluated_model_system_prompt"] = df["evaluated_model_system_prompt"].replace({pd.NA: None})
        if "evaluated_model_retrieved_context" in df.columns:
            df["evaluated_model_retrieved_context"] = df["evaluated_model_retrieved_context"].apply(normalize_context)
            df["evaluated_model_retrieved_context"] = df["evaluated_model_retrieved_context"].astype("object")
        if "evaluated_model_input" in df.columns:
            df["evaluated_model_input"] = df["evaluated_model_input"].astype("string[python]")
            df["evaluated_model_input"] = df["evaluated_model_input"].replace({pd.NA: None})
        if "evaluated_model_output" in df.columns:
            df["evaluated_model_output"] = df["evaluated_model_output"].astype("string[python]")
            df["evaluated_model_output"] = df["evaluated_model_output"].replace({pd.NA: None})
        if "evaluated_model_gold_answer" in df.columns:
            df["evaluated_model_gold_answer"] = df["evaluated_model_gold_answer"].astype("string[python]")
            df["evaluated_model_gold_answer"] = df["evaluated_model_gold_answer"].replace({pd.NA: None})
        if "evaluated_model_attachments" in df.columns:
            df["evaluated_model_attachments"] = df["evaluated_model_attachments"].apply(normalize_attachments)
            df["evaluated_model_attachments"] = df["evaluated_model_attachments"].astype("object")

        # Backfill "dataset_id"
        if dataset_id:
            if "dataset_id" not in df.columns:
                df["dataset_id"] = dataset_id
            else:
                df["dataset_id"] = df["dataset_id"].fillna(dataset_id)

        df = df.sort_values("sid")
        return df


def read_csv(
    filename_or_buffer: Union[str, pathlib.Path, typing.IO[typing.AnyStr]],
    *,
    dataset_id: Optional[str] = None,
    sid_field: str = "sid",
    evaluated_model_system_prompt_field: str = "evaluated_model_system_prompt",
    evaluated_model_retrieved_context_field: str = "evaluated_model_retrieved_context",
    evaluated_model_input_field: str = "evaluated_model_input",
    evaluated_model_output_field: str = "evaluated_model_output",
    evaluated_model_gold_answer_field: str = "evaluated_model_gold_answer",
    **kwargs,
) -> Dataset:
    return _read_dataframe(
        pd.read_csv,
        filename_or_buffer,
        dataset_id=dataset_id,
        sid_field=sid_field,
        evaluated_model_system_prompt_field=evaluated_model_system_prompt_field,
        evaluated_model_retrieved_context_field=evaluated_model_retrieved_context_field,
        evaluated_model_input_field=evaluated_model_input_field,
        evaluated_model_output_field=evaluated_model_output_field,
        evaluated_model_gold_answer_field=evaluated_model_gold_answer_field,
        **kwargs,
    )


def read_jsonl(
    filename_or_buffer,
    *,
    dataset_id: Optional[str] = None,
    sid_field: str = "sid",
    evaluated_model_system_prompt_field: str = "evaluated_model_system_prompt",
    evaluated_model_input_field: str = "evaluated_model_input",
    evaluated_model_retrieved_context_field: str = "evaluated_model_retrieved_context",
    evaluated_model_output_field: str = "evaluated_model_output",
    evaluated_model_gold_answer_field: str = "evaluated_model_gold_answer",
    **kwargs,
) -> Dataset:
    kwargs.setdefault("lines", True)
    return _read_dataframe(
        pd.read_json,
        filename_or_buffer,
        dataset_id=dataset_id,
        sid_field=sid_field,
        evaluated_model_system_prompt_field=evaluated_model_system_prompt_field,
        evaluated_model_retrieved_context_field=evaluated_model_retrieved_context_field,
        evaluated_model_input_field=evaluated_model_input_field,
        evaluated_model_output_field=evaluated_model_output_field,
        evaluated_model_gold_answer_field=evaluated_model_gold_answer_field,
        **kwargs,
    )


def _read_dataframe(
    reader_function,
    filename_or_buffer,
    *,
    dataset_id: Optional[str] = None,
    sid_field: str = "sid",
    evaluated_model_system_prompt_field: str = "evaluated_model_system_prompt",
    evaluated_model_retrieved_context_field: str = "evaluated_model_retrieved_context",
    evaluated_model_input_field: str = "evaluated_model_input",
    evaluated_model_output_field: str = "evaluated_model_output",
    evaluated_model_gold_answer_field: str = "evaluated_model_gold_answer",
    **kwargs,
) -> Dataset:
    df = reader_function(filename_or_buffer, **kwargs)

    if sid_field in df.columns:
        df["sid"] = df[sid_field]
    if evaluated_model_system_prompt_field in df.columns:
        df["evaluated_model_system_prompt"] = df[evaluated_model_system_prompt_field]
    if evaluated_model_retrieved_context_field in df.columns:
        df["evaluated_model_retrieved_context"] = df[evaluated_model_retrieved_context_field]
    if evaluated_model_input_field in df.columns:
        df["evaluated_model_input"] = df[evaluated_model_input_field]
    if evaluated_model_output_field in df.columns:
        df["evaluated_model_output"] = df[evaluated_model_output_field]
    if evaluated_model_gold_answer_field in df.columns:
        df["evaluated_model_gold_answer"] = df[evaluated_model_gold_answer_field]

    dataset_id = _sanitize_dataset_id(dataset_id)
    return Dataset.from_dataframe(df, dataset_id=dataset_id)


def _sanitize_dataset_id(dataset_id: str) -> Optional[str]:
    if not dataset_id:
        return None
    dataset_id = re.sub(r"[^a-zA-Z0-9\-_]", "-", dataset_id.strip())
    if not dataset_id:
        return None
    return dataset_id


class DatasetLoader:
    def __init__(self, loader: typing.Awaitable[Dataset]):
        self.__lock = asyncio.Lock()
        self.__loader = loader
        self.dataset: Optional[Dataset] = None

    async def load(self) -> Dataset:
        async with self.__lock:
            if self.dataset is not None:
                return self.dataset
            self.dataset = await self.__loader
            return self.dataset



import abc
import asyncio
import inspect
import typing
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Union

import typing_extensions

from . import types
from .async_utils import run_as_coro
from .datasets import Row
from .types import _EvalParent

TASK_ARGS = {
    "row",
    "evaluated_model_system_prompt",
    "evaluated_model_retrieved_context",
    "evaluated_model_input",
    "evaluated_model_output",
    "evaluated_model_gold_answer",
    "tags",
    "parent",
}

TaskRetT = Union[types.TaskResult, str]
TaskFn = typing.Callable[..., Union[TaskRetT, typing.Awaitable[TaskRetT]]]


class Task(abc.ABC):
    name: str = ""

    def __init__(self, *, name: str, accepted_args: set[str]):
        self.name = self.name or name or self.__class__.__name__
        self.accepted_args = accepted_args

    def __repr__(self):
        return f"<Task with name {self.name!r} of class {self.__class__.__name__}>"

    async def execute(
        self,
        loop: asyncio.AbstractEventLoop,
        executor: ThreadPoolExecutor,
        row: Row,
        tags: dict[str, str],
        parent: _EvalParent,
    ) -> Optional[types.TaskResult]:
        kwargs = {
            "row": row,
            "evaluated_model_system_prompt": row.evaluated_model_system_prompt,
            "evaluated_model_retrieved_context": row.evaluated_model_retrieved_context,
            "evaluated_model_input": row.evaluated_model_input,
            "evaluated_model_output": row.evaluated_model_output,
            "evaluated_model_gold_answer": row.evaluated_model_gold_answer,
            "tags": {**tags},
            "parent": parent,
        }
        pass_kwargs = {k: v for k, v in kwargs.items() if k in self.accepted_args}
        result = await run_as_coro(
            __loop=loop,
            __executor=executor,
            __fn=self.task,
            **pass_kwargs,
        )
        if result is None:
            return None
        if isinstance(result, str):
            return types.TaskResult(evaluated_model_output=result)
        if isinstance(result, types.TaskResult):
            return result

        raise ValueError(f"Task {self.name!r} returned an object of unexpected type {type(result)!r}")

    @abc.abstractmethod
    def task(self, **kwargs) -> Union[TaskRetT, typing.Awaitable[TaskRetT]]: ...


class FunctionalTask(Task):
    fn: TaskFn

    def __init__(self, *, name: Optional[str] = None, fn: TaskFn, accepted_args: set[str]):
        self.fn = fn
        super().__init__(name=name or fn.__name__, accepted_args=accepted_args)

    async def task(self, **kwargs) -> TaskRetT:
        return await self.fn(**kwargs)


class SyncFunctionalTask(Task):
    fn: TaskFn

    def __init__(self, *, name: Optional[str] = None, fn: TaskFn, accepted_args: set[str]):
        self.fn = fn
        super().__init__(name=name or fn.__name__, accepted_args=accepted_args)

    def task(self, **kwargs) -> TaskRetT:
        return self.fn(**kwargs)


def task(func=None, *, name: Optional[str] = None) -> Union[Task, typing.Callable[[...], Task]]:
    def decorator(fn: TaskFn) -> Task:
        sig = inspect.signature(fn)
        param_keys = sig.parameters.keys()
        for param in param_keys:
            if param not in TASK_ARGS:
                raise ValueError(f"{param!r} is not a valid task argument. Valid arguments are: {TASK_ARGS}")
        if inspect.iscoroutinefunction(fn):
            return FunctionalTask(name=name, fn=fn, accepted_args=set(param_keys))
        else:
            return SyncFunctionalTask(name=name, fn=fn, accepted_args=set(param_keys))

    if func is None:
        return decorator
    return decorator(func)


@typing_extensions.deprecated(
    "'simple_task' is deprecated and will be removed in future versions. Use '@task' decorator instead."
)
def simple_task(lambda_fn: typing.Callable[[str], str]) -> Task:
    @task
    def wrapper(evaluated_model_input: str) -> str:
        return lambda_fn(evaluated_model_input)

    return wrapper


@task
@typing_extensions.deprecated(
    "'nop_task' is deprecated and will be removed in future versions.",
)
def nop_task(evaluated_model_output: str) -> str:
    return evaluated_model_output



import datetime
import re
import typing
from typing import Optional, Union

import pydantic
from typing_extensions import Annotated


class Account(pydantic.BaseModel):
    id: str
    name: str


class WhoAmIAPIKey(pydantic.BaseModel):
    id: str
    account: Account


class WhoAmICaller(pydantic.BaseModel):
    api_key: WhoAmIAPIKey


class WhoAmIResponse(pydantic.BaseModel):
    caller: WhoAmICaller


class Evaluator(pydantic.BaseModel):
    id: str
    name: str
    evaluator_family: Optional[str]
    aliases: Optional[list[str]]


class ListEvaluatorsResponse(pydantic.BaseModel):
    evaluators: list[Evaluator]


class Project(pydantic.BaseModel):
    id: str
    name: str


class CreateProjectRequest(pydantic.BaseModel):
    name: str


class GetProjectResponse(pydantic.BaseModel):
    project: Project


class Experiment(pydantic.BaseModel):
    project_id: str
    id: str
    name: str
    tags: Optional[dict[str, str]] = None


class CreateExperimentRequest(pydantic.BaseModel):
    project_id: str
    name: str
    tags: dict[str, str] = pydantic.Field(default_factory=dict)


class CreateExperimentResponse(pydantic.BaseModel):
    experiment: Experiment


class GetExperimentResponse(pydantic.BaseModel):
    experiment: Experiment


class EvaluateEvaluator(pydantic.BaseModel):
    evaluator: str
    criteria: Optional[str] = None
    explain_strategy: str = "always"


class EvaluatedModelAttachment(pydantic.BaseModel):
    url: str
    media_type: str
    usage_type: Optional[str] = "evaluated_model_input"


# See https://docs.patronus.ai/reference/evaluate_v1_evaluate_post for request field descriptions.
class EvaluateRequest(pydantic.BaseModel):
    evaluators: list[EvaluateEvaluator] = pydantic.Field(min_length=1)
    evaluated_model_system_prompt: Optional[str] = None
    evaluated_model_retrieved_context: Optional[Union[list[str], str]] = None
    evaluated_model_input: Optional[str] = None
    evaluated_model_output: Optional[str] = None
    evaluated_model_gold_answer: Optional[str] = None
    evaluated_model_attachments: Optional[list[EvaluatedModelAttachment]] = None
    app: Optional[str] = None
    experiment_id: Optional[str] = None
    capture: str = "all"
    dataset_id: Optional[str] = None
    dataset_sample_id: Optional[int] = None
    tags: Optional[dict[str, str]] = None


class EvaluationResultAdditionalInfo(pydantic.BaseModel):
    positions: Optional[list]
    extra: Optional[dict]
    confidence_interval: Optional[dict]


class EvaluationResult(pydantic.BaseModel):
    id: Optional[str] = None
    project_id: Optional[str] = None
    app: Optional[str] = None
    experiment_id: Optional[str] = None
    created_at: Optional[pydantic.AwareDatetime] = None
    evaluator_id: str
    criteria: str
    evaluated_model_system_prompt: Optional[str] = None
    evaluated_model_retrieved_context: Optional[list[str]] = None
    evaluated_model_input: Optional[str] = None
    evaluated_model_output: Optional[str] = None
    evaluated_model_gold_answer: Optional[str] = None
    pass_: Optional[bool] = pydantic.Field(default=None, alias="pass")
    score_raw: Optional[float] = None
    text_output: Optional[str] = None
    additional_info: Optional[EvaluationResultAdditionalInfo] = None
    evaluation_metadata: Optional[dict] = None
    explanation: Optional[str] = None
    evaluation_duration: Optional[datetime.timedelta] = None
    explanation_duration: Optional[datetime.timedelta] = None
    evaluator_family: str
    evaluator_profile_public_id: str
    dataset_id: Optional[str] = None
    dataset_sample_id: Optional[int] = None
    tags: Optional[dict[str, str]] = None


class EvaluateResult(pydantic.BaseModel):
    evaluator_id: str
    criteria: str
    status: str
    error_message: Optional[str]
    evaluation_result: Optional[EvaluationResult]


class EvaluateResponse(pydantic.BaseModel):
    results: list[EvaluateResult]


def sanitize_evaluator_id(v: typing.Any, info: pydantic.ValidationInfo):
    if not isinstance(v, str):
        return v
    v = v.strip()
    return re.sub(r"[^a-zA-Z0-9\-_./]", "-", v)


class ExportEvaluationResult(pydantic.BaseModel):
    app: Optional[str] = None
    experiment_id: Optional[str] = None
    evaluator_id: Annotated[str, pydantic.BeforeValidator(sanitize_evaluator_id)]
    criteria: Optional[str] = None
    evaluated_model_system_prompt: Optional[str] = None
    evaluated_model_retrieved_context: Optional[list[str]] = None
    evaluated_model_input: Optional[str] = None
    evaluated_model_output: Optional[str] = None
    evaluated_model_gold_answer: Optional[str] = None
    evaluated_model_attachments: Optional[list[EvaluatedModelAttachment]] = None
    pass_: Optional[bool] = pydantic.Field(default=None, serialization_alias="pass")
    score_raw: Optional[float] = None
    text_output: Optional[str] = None
    explanation: Optional[str] = None
    evaluation_duration: Optional[datetime.timedelta] = None
    explanation_duration: Optional[datetime.timedelta] = None
    evaluation_metadata: Optional[dict[str, typing.Any]] = None
    evaluated_model_name: Optional[str] = None
    evaluated_model_provider: Optional[str] = None
    evaluated_model_params: Optional[dict[str, Union[str, int, float]]] = None
    evaluated_model_selected_model: Optional[str] = None
    dataset_id: Optional[str] = None
    dataset_sample_id: Optional[int] = None
    tags: Optional[dict[str, str]] = None


class ExportEvaluationRequest(pydantic.BaseModel):
    evaluation_results: list[ExportEvaluationResult]


class ExportEvaluationResultPartial(pydantic.BaseModel):
    id: str
    app: Optional[str]
    created_at: pydantic.AwareDatetime
    evaluator_id: str


class ExportEvaluationResponse(pydantic.BaseModel):
    evaluation_results: list[ExportEvaluationResultPartial]


class ListCriteriaRequest(pydantic.BaseModel):
    public_id: Optional[str] = None
    evaluator_family: Optional[str] = None
    evaluator_id: Optional[str] = None
    name: Optional[str] = None
    revision: Optional[str] = None
    get_last_revision: bool = False
    is_patronus_managed: Optional[bool] = None
    limit: int = 1000
    offset: int = 0


class EvaluatorCriteria(pydantic.BaseModel):
    public_id: str
    evaluator_family: str
    name: str
    revision: int
    config: Optional[dict[str, typing.Any]]
    is_patronus_managed: bool
    created_at: datetime.datetime
    description: Optional[str]


class CreateCriteriaRequest(pydantic.BaseModel):
    evaluator_family: str
    name: str
    config: dict[str, typing.Any]


class CreateCriteriaResponse(pydantic.BaseModel):
    evaluator_criteria: EvaluatorCriteria


class AddEvaluatorCriteriaRevisionRequest(pydantic.BaseModel):
    config: dict[str, typing.Any]


class AddEvaluatorCriteriaRevisionResponse(pydantic.BaseModel):
    evaluator_criteria: EvaluatorCriteria


class ListCriteriaResponse(pydantic.BaseModel):
    evaluator_criteria: list[EvaluatorCriteria]


class DatasetDatum(pydantic.BaseModel):
    dataset_id: str
    sid: int
    evaluated_model_system_prompt: Optional[str] = None
    evaluated_model_retrieved_context: Optional[list[str]] = None
    evaluated_model_input: Optional[str] = None
    evaluated_model_output: Optional[str] = None
    evaluated_model_gold_answer: Optional[str] = None
    meta_evaluated_model_name: Optional[str] = None
    meta_evaluated_model_provider: Optional[str] = None
    meta_evaluated_model_selected_model: Optional[str] = None
    meta_evaluated_model_params: Optional[dict[str, Union[str, int, float]]] = None


class ListDatasetData(pydantic.BaseModel):
    data: list[DatasetDatum]



import asyncio
import threading
from concurrent.futures import Executor, ThreadPoolExecutor
from typing import Callable, Optional, Any


class EvaluationsBackgroundRunner:
    def __init__(self, *, loop: Optional[asyncio.AbstractEventLoop]=None, executor: Optional[Executor]=None):
        self._loop = loop or asyncio.get_event_loop_policy().new_event_loop()
        self._executor = executor or ThreadPoolExecutor()

        self._thread = threading.Thread(
            target=self._run_loop,
            # We don't want to close the thread abruptly, on program exit.
            # When program exists we try to finish pending requests and do cleanup.
            daemon=False,
        )
        self._thread.start()

    def _run_loop(self) -> None:
        asyncio.get_event_loop_policy().set_event_loop(self._loop)
        self._loop.run_forever()
        self._loop.close()

    def submit(self, fn: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
        if asyncio.iscoroutinefunction(fn):
            coro = fn(*args, *kwargs)
            asyncio.run_coroutine_threadsafe(coro, loop=self._loop)
        self._loop.run_in_executor(self._executor, fn, args, kwargs)

    def shutdown(self) -> None:
        self._loop.stop()
        self._thread.join()


import asyncio
import datetime
import inspect
import itertools
import logging
import os
import re
import statistics
import sys
import time
import traceback
import typing
import urllib.parse
import warnings
from concurrent.futures import ThreadPoolExecutor
from sys import version_info
from typing import Optional, Union

import pandas as pd
import typing_extensions as te
from tqdm.asyncio import tqdm as tqdm_async

from . import api_types, types
from .async_utils import run_until_complete
from .client import Client
from .config import config
from .datasets import Dataset, DatasetLoader, Row
from .evaluators import Evaluator
from .evaluators_remote import RemoteEvaluator
from .retry import retry
from .tasks import Task, nop_task

log = logging.getLogger(__name__)


class EvalLink(typing.TypedDict, total=False):
    task: Optional[Task]
    evaluators: list[Evaluator]


class AsyncTQDMWithHandle(tqdm_async):
    """
    Workaround for accessing tqdm instance with async tasks.
    Instead of calling gather which don't provide access to tqdm instance:
    ```
    tqdm_async.gather(features)
    ```

    Call prep_gather() follow by gather()
    ```
    tqdm_instance = AsyncTQDMWithHandle.pre_gather(features)
    ...
    tqdm_instance.gather()
    ```

    tqdm_instance can be used to clear and display progress bar using tqdm_instance.clear() and tqdm_instance.display()
    methods.
    """

    async def gather(self):
        def into_iter():
            yield from self

        res = [await f for f in into_iter()]
        return [i for _, i in sorted(res)]

    @classmethod
    async def prep_gather(cls, *fs, loop=None, timeout=None, total=None, **tqdm_kwargs) -> te.Self:
        async def wrap_awaitable(i, f):
            return i, await f

        ifs = [wrap_awaitable(i, f) for i, f in enumerate(fs)]
        return cls.prep_as_completed(ifs, loop=loop, timeout=timeout, total=total, **tqdm_kwargs)

    @classmethod
    def prep_as_completed(cls, fs, *, loop=None, timeout=None, total=None, **tqdm_kwargs):
        if total is None:
            total = len(fs)
        kwargs = {}
        if version_info[:2] < (3, 10):
            kwargs["loop"] = loop
        return cls(
            asyncio.as_completed(fs, timeout=timeout, **kwargs),
            total=total,
            **tqdm_kwargs,
        )


class ReportedTaskResult(typing.NamedTuple):
    sid: int
    link_idx: int
    task_name: str
    task_result: types.TaskResult


class ReportedEvaluationResult(typing.NamedTuple):
    sid: int
    link_idx: int
    evaluator_id: str
    criteria: Optional[str]
    evaluation_result: api_types.ExportEvaluationResult
    evaluation_metadata: Optional[dict[str, typing.Any]]
    evaluation_explanation: Optional[str]


class Reporter:
    # batch_size 10 is max batch size that Patronus AI API accepts in single call
    batch_size = 10
    tqdm = None

    def __init__(self, client: Client, experiment_id: str, flush_interval: int = 10):
        self._client = client
        self.experiment_id = experiment_id

        self._lock = asyncio.Lock()
        self.last_export_slice = slice(0, 0)
        self.last_flush_ts = time.time()
        self.flush_interval = flush_interval

        self.task_results: list[ReportedTaskResult] = []

        EvaluatorIDT = str
        CriteriaT = str
        self.evaluators: set[tuple[EvaluatorIDT, CriteriaT]] = set()
        self.remote_results: list[ReportedEvaluationResult] = []
        self.outgoing_results: list[ReportedEvaluationResult] = []

        self.task_errors = []
        self.evaluator_errors = []

    def set_tqdm(self, tqdm: AsyncTQDMWithHandle):
        self.tqdm = tqdm

    def add_task_result(self, sid: int, link_idx: int, task_name: str, r: types.TaskResult):
        self.task_results.append(ReportedTaskResult(sid=sid, link_idx=link_idx, task_name=task_name, task_result=r))

    async def add_evaluation_result(
        self,
        evaluator_name: str,
        criteria: Optional[str],
        link_idx: int,
        row: Row,
        task_result: Optional[types.TaskResult],
        evaluation_result: types.EvaluatorOutput,
        tags: dict[str, str],
        *,
        dataset_id: str,
        dataset_sample_id: int,
        already_captured: bool,
    ):
        model_output = (task_result and task_result.evaluated_model_output) or row.evaluated_model_output
        task_metadata = (task_result and task_result.metadata) or {}

        eval_metadata = {}
        evaluation_duration = None
        explanation_duration = None
        if isinstance(evaluation_result.result, types.EvaluationResult):
            eval_metadata = evaluation_result.result.metadata
            edur = evaluation_result.result.evaluation_duration_s
            if not edur:
                edur = evaluation_result.duration
            evaluation_duration = datetime.timedelta(seconds=edur)
            if evaluation_result.result.explanation_duration_s:
                explanation_duration = datetime.timedelta(seconds=evaluation_result.result.explanation_duration_s)
        elif isinstance(evaluation_result.result, api_types.EvaluationResult):
            eval_metadata = evaluation_result.result.evaluation_metadata
            evaluation_duration = evaluation_result.result.evaluation_duration
            explanation_duration = evaluation_result.result.explanation_duration

        entry = ReportedEvaluationResult(
            sid=dataset_sample_id,
            link_idx=link_idx,
            evaluator_id=evaluator_name,
            criteria=criteria,
            evaluation_result=api_types.ExportEvaluationResult(
                experiment_id=self.experiment_id,
                evaluator_id=evaluator_name,
                criteria=criteria,
                evaluated_model_system_prompt=row.evaluated_model_system_prompt,
                evaluated_model_retrieved_context=row.evaluated_model_retrieved_context,
                evaluated_model_input=row.evaluated_model_input,
                evaluated_model_output=model_output,
                evaluated_model_gold_answer=row.evaluated_model_gold_answer,
                evaluated_model_attachments=row.evaluated_model_attachments,
                pass_=evaluation_result.result.pass_,
                score_raw=evaluation_result.result.score_raw,
                text_output=evaluation_result.result.text_output,
                explanation=evaluation_result.result.explanation,
                evaluation_duration=evaluation_duration,
                explanation_duration=explanation_duration,
                evaluation_metadata=eval_metadata,
                evaluated_model_name=task_metadata.get("evaluated_model_name"),
                evaluated_model_provider=task_metadata.get("evaluated_model_provider"),
                evaluated_model_params=task_metadata.get("evaluated_model_params"),
                evaluated_model_selected_model=task_metadata.get("evaluated_model_selected_model"),
                dataset_id=dataset_id,
                dataset_sample_id=dataset_sample_id,
                tags=tags,
            ),
            evaluation_metadata=eval_metadata,
            evaluation_explanation=evaluation_result.result.explanation,
        )

        async with self._lock:
            self.evaluators.add((evaluator_name, criteria))
            if already_captured:
                self.remote_results.append(entry)
            else:
                self.outgoing_results.append(entry)

        await self._conditional_flush()

    async def _conditional_flush(self):
        async with self._lock:
            buffered = len(self.outgoing_results) - self.last_export_slice.stop
            if buffered == 0:
                return
            if buffered >= self.batch_size or self.last_flush_ts + self.flush_interval < time.time():
                await self._flush()

    async def flush(self):
        async with self._lock:
            buffered = len(self.outgoing_results) - self.last_export_slice.stop
            if buffered == 0:
                return
            await self._flush()

    async def _flush(self):
        while self.last_export_slice.stop < len(self.outgoing_results):
            upper_idx = min(
                len(self.outgoing_results),
                self.last_export_slice.stop + self.batch_size,
            )
            self.last_export_slice = slice(self.last_export_slice.stop, upper_idx)
            results_to_export = self.outgoing_results[self.last_export_slice]

            @retry(max_attempts=5, initial_delay=2)
            async def call():
                await self._client.api.export_evaluations(
                    api_types.ExportEvaluationRequest(
                        evaluation_results=list(e.evaluation_result for e in results_to_export),
                    )
                )
                log.debug(f"Exported {len(results_to_export)} results")

            await call()
        self.last_flush_ts = time.time()

    async def task_error(self, err: Exception, datum, dataset_data_id):
        stack_trace = getattr(err, "stack_trace", None)
        if stack_trace is None:
            stack_trace = traceback.format_exc()
        self.print_error(stack_trace)
        self.print_error(f"Task failed on sample {dataset_data_id!r} with the  following error: {err}")
        self.task_errors.append((err, datum, dataset_data_id))

    async def evaluator_error(self, err: Exception, datum, dataset_data_id, evaluator: str, criteria: str):
        stack_trace = getattr(err, "stack_trace", None)
        if stack_trace is None:
            stack_trace = traceback.format_exc()
        self.print_error(stack_trace)
        self.print_error(
            f"Evaluator ({evaluator}, {criteria}) failed on sample {dataset_data_id} with the following error: {err}"
        )
        self.evaluator_errors.append((err, datum, dataset_data_id, evaluator, criteria))

    def print_error(self, message: str):
        if self.tqdm:
            self.tqdm.clear()
        print(message, file=sys.stderr)
        if self.tqdm:
            self.tqdm.display()

    def summary(self):
        for evaluator_name, criteria in self.evaluators:
            name = evaluator_name if not criteria else f"{evaluator_name}:{criteria}"
            results: typing.Iterator[ReportedEvaluationResult] = itertools.chain(
                self.outgoing_results, self.remote_results
            )
            results = filter(
                lambda r: r.evaluation_result.evaluator_id == evaluator_name
                and r.evaluation_result.criteria == criteria,
                results,
            )
            scores_and_passes = list(
                map(
                    lambda r: (
                        r.evaluation_result.score_raw,
                        r.evaluation_result.pass_,
                    ),
                    results,
                )
            )

            scores = [x[0] for x in scores_and_passes if x[0] is not None]
            passes = [int(x[1]) for x in scores_and_passes if x[1] is not None]

            print_summary(name, scores, passes, len(scores_and_passes), display_hist=True)

        if self.task_errors or self.evaluator_errors:
            print()
        if self.task_errors:
            print(f"Task failures: {len(self.task_errors)}")
        if self.evaluator_errors:
            print(f"Evaluators failures: {len(self.evaluator_errors)}")

    def task_results_df(self) -> pd.DataFrame:
        if not self.task_results:
            columns = [
                "sid",
                "link_idx",
                "task_name",
                "task_result_evaluated_model_output",
                "task_result_metadata",
            ]
            return pd.DataFrame(columns=columns).set_index(["sid", "link_idx"])

        def dict_with_sid(r: ReportedTaskResult) -> dict:
            return {
                "sid": r.sid,
                "link_idx": r.link_idx,
                "task_name": r.task_name,
                "task_result_evaluated_model_output": r.task_result.evaluated_model_output,
                "task_result_metadata": r.task_result.metadata,
            }

        df = pd.DataFrame.from_records(map(dict_with_sid, self.task_results)).sort_values(
            ["sid", "link_idx", "task_name"]
        )
        df.set_index(["sid", "link_idx"], inplace=True)
        return df

    def evaluation_results_df(self):
        if not self.remote_results and not self.outgoing_results:
            return pd.DataFrame()
        all_eval_results = itertools.chain(self.remote_results, self.outgoing_results)

        def mapper(r: ReportedEvaluationResult) -> dict:
            return {
                "sid": r.sid,
                "link_idx": r.link_idx,
                "evaluator_id": r.evaluator_id,
                "criteria": r.criteria,
                "pass": r.evaluation_result.pass_,
                "score_raw": r.evaluation_result.score_raw,
                "evaluation_duration": r.evaluation_result.evaluation_duration,
                "evaluation_metadata": r.evaluation_metadata,
                "evaluation_explanation": r.evaluation_explanation,
                "tags": r.evaluation_result.tags,
            }

        df = pd.DataFrame.from_records(map(mapper, all_eval_results))
        if df.empty:
            return df
        df = df.sort_values(["sid", "link_idx", "evaluator_id", "criteria"])
        df.set_index("sid", inplace=True)
        return df


async def with_semaphore(sem, coro):
    async with sem:
        return await coro


class Experiment:
    project_name: str
    project_id: Optional[str]
    experiment_id: Optional[str]
    # TODO handle with API?
    experiment_name: str
    whoami: Optional[api_types.WhoAmIResponse]

    __raw_dataset: typing.Any
    dataset: Optional[Dataset]
    chain: list[EvalLink]

    tags: dict[str, str]

    reporter: Optional[Reporter]

    _client: Optional[Client]

    _pool: ThreadPoolExecutor
    _sem: asyncio.Semaphore

    def __init__(
        self,
        client: Optional[Client],
        project_name: Optional[str],
        dataset,  # TODO type hint
        chain: list[EvalLink],
        tags: Optional[dict[str, str]],
        max_concurrency: int,
        experiment_name: str = "",
        experiment_id: Optional[str] = None,
        **kwargs,
    ):
        self._client = client
        self.project_id = None
        self.project_name = re.sub(r"[^a-zA-Z0-9\-_ ]", "-", project_name.strip()) or "default"
        self.experiment_id = experiment_id
        self.experiment_name = generate_experiment_name(experiment_name)

        self.__raw_dataset = dataset
        self.dataset = None
        self.chain = chain

        self.tags = tags or {}

        self._sem = asyncio.Semaphore(max_concurrency)
        self._pool = ThreadPoolExecutor()

        self.reporter = None

    async def prepare(self):
        for eval_link in self.chain:
            if (task := eval_link.get("task")) and not isinstance(task, Task):
                raise ValueError(f"task {task!r} must inherit from Task. Did you forget to use @task decorator?")

        print("Preparing dataset... ", end="")
        self.dataset = await self.fetch_dataset(self.__raw_dataset)
        print("DONE")

        print("Preparing evaluators... ", end="")
        for eval_link in self.chain:
            evaluators = eval_link.get("evaluators", [])
            for e in evaluators:
                if not isinstance(e, Evaluator):
                    raise ValueError(
                        f"evaluator {e!r} must inherit from Evaluator. Did you forget to use @evaluator decorator?"
                    )

            # Preload all remote evaluators
            preloads = [e.load() for e in evaluators if isinstance(e, RemoteEvaluator)]
            if preloads:
                await asyncio.gather(*preloads)

        print("DONE")

        if not self._client:
            return

        self.whoami = await self._client.api.whoami()

        if not self.experiment_id:
            project = await self._client.api.create_project(api_types.CreateProjectRequest(name=self.project_name))
            self.project_id = project.id
            self.project_name = project.name

            ex = await self._client.api.create_experiment(
                api_types.CreateExperimentRequest(project_id=self.project_id, name=self.experiment_name, tags=self.tags)
            )
            self.experiment_id = ex.id
        else:
            ex = await self._client.api.get_experiment(self.experiment_id)
            if ex is None:
                raise ValueError(f"experiment with id {self.experiment_id!r} not found")
            project = await self._client.api.get_project(ex.project_id)
            self.experiment_name = ex.name
            self.project_id = project.id
            self.project_name = project.name

        # TODO associate reported that doesn't need client if client not available
        self.reporter = Reporter(self._client, self.experiment_id, flush_interval=10)

    @classmethod
    async def fetch_dataset(cls, dataset) -> Dataset:
        if isinstance(dataset, Dataset):
            return dataset
        elif isinstance(dataset, DatasetLoader):
            return await dataset.load()
        elif isinstance(dataset, (list, tuple)):
            return Dataset.from_records(dataset)
        elif isinstance(dataset, pd.DataFrame):
            return Dataset.from_dataframe(dataset)
        elif inspect.iscoroutine(dataset):
            return await cls.fetch_dataset(await dataset)
        elif inspect.iscoroutinefunction(dataset):
            return await cls.fetch_dataset(await dataset())
        elif callable(dataset):
            return await cls.fetch_dataset(dataset())
        else:
            raise ValueError(f"'dataset' passed to the experiment is an unexpected object of type {type(dataset)!r}")

    async def run(self):
        title = f"Experiment  {self.project_name}/{self.experiment_name}"
        print("=" * len(title))

        tasks = []
        for row in self.dataset.iterrows():
            task = self.run_task_and_eval(row, dataset_id=self.dataset.dataset_id, dataset_sample_id=row.sid)
            tasks.append(task)

        tqdm = await AsyncTQDMWithHandle.prep_gather(*tasks, desc=title, unit="sample")
        self.reporter.set_tqdm(tqdm)
        await tqdm.gather()

        await self.reporter.flush()

        self.reporter.summary()

        print()
        print(get_link(self.whoami.caller.api_key.account.id, self.experiment_id))

    async def run_task_and_eval(self, row: Row, dataset_id: Optional[str], dataset_sample_id: int):
        loop = asyncio.get_running_loop()

        parent = types._EvalParent(task=None, evals=None, parent=None)

        for link_idx, eval_link in enumerate(self.chain):
            task = eval_link.get("task")
            evaluators = eval_link.get("evaluators", [])

            outgoing_tags = self.tags
            task_result = None

            if task:
                try:
                    task_result = await task.execute(
                        loop,
                        self._pool,
                        row=row,
                        tags=self.tags,
                        parent=parent,
                    )
                except Exception as e:
                    await self.reporter.task_error(e, row, dataset_sample_id)
                    return

                if task_result is None:
                    parent = types._EvalParent(task=None, evals=None, parent=parent)
                    # If task returned non it means the record processing should be skipped
                    return

                self.reporter.add_task_result(dataset_sample_id, link_idx, task.name, task_result)

                if task_result.tags:
                    outgoing_tags = merge_tags({}, task_result.tags, experiment_tags=self.tags)

            futures = [
                loop.create_task(
                    with_semaphore(
                        self._sem,
                        evaluator.execute(
                            loop=loop,
                            executor=self._pool,
                            row=row,
                            task_result=task_result,
                            tags=outgoing_tags,
                            parent=parent,
                            experiment_id=self.experiment_id,
                            dataset_id=dataset_id,
                            dataset_sample_id=dataset_sample_id,
                        ),
                    )
                )
                for evaluator in evaluators
            ]

            has_eval_errors = False
            eval_results_map = types.EvalsMap()

            for evaluator, f in zip(evaluators, futures):
                try:
                    eval_result: Optional[types.EvaluatorOutput] = await f
                except Exception as e:
                    await self.reporter.evaluator_error(
                        e,
                        row,
                        dataset_sample_id,
                        evaluator.name,
                        evaluator.criteria,
                    )
                    eval_results_map[evaluator.display_name()] = None
                    has_eval_errors = True
                    continue

                eval_results_map[evaluator.display_name()] = eval_result.result if eval_result else None
                if eval_result is None:
                    # If evaluator returned None it means the evaluation has been skipped
                    continue

                eval_tags = eval_result.result.tags or {}
                outgoing_tags = merge_tags(outgoing_tags, eval_tags, experiment_tags=self.tags)

                await self.reporter.add_evaluation_result(
                    evaluator.name,
                    evaluator.criteria,
                    link_idx,
                    row,
                    task_result,
                    eval_result,
                    outgoing_tags,
                    dataset_id=dataset_id,
                    dataset_sample_id=dataset_sample_id,
                    already_captured=evaluator.remote_capture,
                )

            if has_eval_errors:
                return
            parent = types._EvalParent(task=task_result, evals=eval_results_map, parent=parent)

    def to_dataframe(self) -> pd.DataFrame:
        df_dataset = self.dataset.df.set_index("sid", inplace=False)
        df_task_results = self.reporter.task_results_df()
        df_evaluation_results = self.reporter.evaluation_results_df()
        df = df_evaluation_results.join(df_task_results, on=["sid", "link_idx"], how="left")
        df = df.join(df_dataset, on="sid", how="left")
        return df

    def to_dataset(
        self,
        *,
        dataset_id: Optional[str] = None,
        rename_columns: Optional[dict[str, str]] = None,
    ) -> Dataset:
        df = self.to_dataframe()
        if rename_columns:
            col_to_drop = [col for col in rename_columns.keys() if col in df.columns]
            df.drop(columns=col_to_drop, inplace=True)
            df.rename(columns=rename_columns, inplace=True)
        return Dataset.from_dataframe(df, dataset_id=dataset_id)

    def to_csv(self, path_or_buf, sep: str = ",", **kwargs) -> Optional[str]:
        df = self.to_dataframe()
        return df.to_csv(path_or_buf, sep, **kwargs)


def experiment(
    client: Optional[Client],
    project_name: Optional[str],
    *,
    dataset,
    task: Optional[Task] = None,
    evaluators: Optional[list[Evaluator]] = None,
    #
    chain: Optional[list[EvalLink]] = None,
    tags: Optional[dict[str, str]] = None,
    experiment_name: str = "",
    max_concurrency: int = 10,
    **kwargs,
) -> Union[typing.Awaitable[Experiment], Experiment]:
    if chain:
        assert (
            task is nop_task and not evaluators
        ), "Passing 'chain' argument prevents use of 'task' and 'evaluators' arguments."
    if dataset is None:
        dataset = kwargs.get("data")

    if not chain:
        chain = [{"task": task, "evaluators": evaluators}]

    ex = Experiment(
        client=client,
        project_name=project_name,
        dataset=dataset,
        chain=chain,
        tags=tags or {},
        max_concurrency=max_concurrency,
        experiment_name=experiment_name,
        **kwargs,
    )

    async def run() -> Experiment:
        await ex.prepare()
        await ex.run()
        return ex

    return run_until_complete(run())


def print_summary(name: str, scores: list[float], passes: list[int], count: int, display_hist: bool):
    title = f"Summary: {name}"

    print()
    print(title)
    print("-" * len(title))
    print(f"Count     : {count}")
    print(f"Pass rate : {round(statistics.mean(passes), 3)}")
    print(f"Mean      : {round(statistics.mean(scores), 3)}")
    print(f"Min       : {round(min(scores), 3)}")
    print(f"25%       : {round(percentile(scores, 25), 3)}")
    print(f"50%       : {round(percentile(scores, 50), 3)}")
    print(f"75%       : {round(percentile(scores, 75), 3)}")
    print(f"Max       : {round(max(scores), 3)}")

    if display_hist:
        print()
        print("Score distribution")
        print_histogram(scores)


def gen_name(name: str) -> str:
    name = re.sub(r"[^a-zA-Z0-9]", "", name).lower()
    ts = datetime.datetime.now().strftime("%y%m%d%H%M%S")
    name = name or "unknown"
    return f"ex-{name}-{ts}"


def percentile(data: list[float], p: int):
    data = sorted(data)
    index = (p / 100) * (len(data) - 1)
    if index.is_integer():
        return data[int(index)]
    else:
        lower_bound = int(index)
        upper_bound = lower_bound + 1
        weight = index - lower_bound
        return data[lower_bound] * (1 - weight) + data[upper_bound] * weight


def print_histogram(data, bin_count=5):
    # Calculate the range of the data
    min_val = min(data)
    max_val = max(data)

    if min_val == max_val:
        if min_val > 0.5:
            min_val = 0
        else:
            max_val = 1

    range_val = max_val - min_val

    # Calculate bin size
    bin_size = range_val / bin_count

    # Initialize bins
    bins = [0] * bin_count

    # Distribute data into bins
    for value in data:
        # Find the appropriate bin for the current value
        bin_index = int((value - min_val) / bin_size)
        # Edge case for the maximum value
        if bin_index == bin_count:
            bin_index -= 1
        bins[bin_index] += 1

    # Determine the width of the histogram
    max_bin_count = max(bins)
    scale_factor = 20 / max_bin_count  # Scale the histogram to a max width of 50 characters

    # Print the histogram
    print("Score Range".ljust(20), "Count".ljust(10), "Histogram")
    for i in range(bin_count):
        bin_start = min_val + i * bin_size
        bin_end = bin_start + bin_size
        bin_count = bins[i]
        bar = "#" * int(bin_count * scale_factor)
        print(f"{bin_start:.2f} - {bin_end:.2f}".ljust(20), f"{bin_count}".ljust(10), bar)


def get_link(account_id: str, experiment_id: str) -> str:
    params = {"account_id": account_id}
    ui_url = config().ui_url.rstrip("/")
    return f"{ui_url}/experiments/{experiment_id}?{urllib.parse.urlencode(params)}"


def generate_experiment_name(name: str) -> str:
    ts = int(time.time())
    if name:
        return f"{name}-{ts}"
    try:
        login = os.getlogin()
        return f"{login}-{ts}"
    except OSError:  # Possible in-cluster error: No such device or address
        return str(ts)


def merge_tags(tags: dict, new_tags: dict, experiment_tags: dict) -> dict:
    tags = {**tags, **new_tags}
    common_keys = set(tags.keys()) & set(experiment_tags.keys())
    diff = {key for key in common_keys if tags[key] != experiment_tags[key]}
    if diff:
        warnings.warn(f"Overriding experiment tags is not allowed. Tried to override tag: {list(diff)!r}")
    return {**tags, **experiment_tags}



from .api_types import EvaluateRequest as EvaluateRequest
from .client import Client as Client
from .datasets import Dataset as Dataset
from .datasets import Row as Row
from .datasets import read_csv as read_csv
from .datasets import read_jsonl as read_jsonl
from .evaluators import Evaluator as Evaluator
from .evaluators import evaluator as evaluator
from .evaluators import simple_evaluator as simple_evaluator
from .retry import retry as retry
from .tasks import Task as Task
from .tasks import nop_task as nop_task
from .tasks import simple_task as simple_task
from .tasks import task as task
from .types import EvalParent as EvalParent
from .types import EvaluationResult as EvaluationResult
from .types import TaskResult as TaskResult



import atexit
import json
import random
import traceback
import typing
import functools
import time
import uuid

import opentelemetry
from opentelemetry import trace
from opentelemetry._logs import SeverityNumber
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.trace import TracerProvider

from opentelemetry.trace import TraceFlags

from opentelemetry import _logs
from opentelemetry.sdk import _logs as logs
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.trace.export import BatchSpanProcessor

logger_provider = logs.LoggerProvider()
_logs.set_logger_provider(logger_provider)

app = "non-default"
headers = {"pat-project-name": "Hello", "x-api-key": "sk-FgvQiD_0UgJMPqyuX3TcBnddSI7IBApguZ3VTcg9hvo"}

exporter = OTLPLogExporter(headers=headers)
logger_provider.add_log_record_processor(BatchLogRecordProcessor(exporter))
handler = logs.LoggingHandler(logger_provider=logger_provider)

texp = OTLPSpanExporter(headers=headers)
tprov = TracerProvider()
tprov.add_span_processor(BatchSpanProcessor(texp))

atexit.register(logger_provider.shutdown)
atexit.register(tprov.shutdown)

# logger = _logs.get_logger("patronus")
logger = logger_provider.get_logger("patronus")
tracer = tprov.get_tracer("patronus")

def jsonify(v: typing.Any):
    try:
        return json.loads(json.dumps(v))
    except TypeError:
        return str(v)


def clean_dict(d: dict):
    if isinstance(d, dict):
        keys_to_delete = []
        for k, v in d.items():
            if v is None:
                keys_to_delete.append(k)
            else:
                clean_dict(v)
        for k in keys_to_delete:
            del d[k]
    if isinstance(d, list):
        for item in d:
            clean_dict(item)


def traced(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with tracer.start_as_current_span(func.__name__) as span:

            # TODO remove
            time.sleep(random.random()/10)

            attributes = {
                # "pat.app": app,
                "pat.experiment.id": str(uuid.uuid4()),
                "function_name": func.__name__
            }

            span: trace.Span
            span.set_attributes(attributes)
            span_ctx = span.get_span_context()
            ret = None
            exc = None
            exc_info = None
            try:
                ret = func(*args, **kwargs)
            except Exception as e:
                exc = e
                exc_info = traceback.format_exc()

            sev = SeverityNumber.ERROR if exc else SeverityNumber.INFO
            body = {
                "input": {
                    **{f"{i}": jsonify(arg) for i, arg in enumerate(args)},
                    **{k: jsonify(v) for k, v in kwargs.items()},
                },
                "output": jsonify(ret) or {},
            }
            if exc is not None:
                body["exception"] = str(exc)
                body["stack_trace"] = exc_info

            log = logs.LogRecord(
                timestamp=time.time_ns(),
                trace_flags=TraceFlags.SAMPLED,
                trace_id=span_ctx.trace_id,
                span_id=span_ctx.span_id,
                severity_text=sev.name,
                severity_number=sev,
                body=body,
                attributes={"pat.log_type": "function_call", **attributes},
            )
            # print(log.to_json())
            logger.emit(log)

            if exc:
                raise exc
            return ret

    return wrapper


import functools
import typing
from typing import Optional

import pydantic
import pydantic_settings


class Config(pydantic_settings.BaseSettings):
    model_config = pydantic_settings.SettingsConfigDict(env_prefix="patronus_", yaml_file="patronus.yaml")

    api_key: Optional[str] = pydantic.Field(None)
    api_url: str = pydantic.Field("https://api.patronus.ai")
    ui_url: str = pydantic.Field("https://app.patronus.ai")

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: typing.Type[pydantic_settings.BaseSettings],
        init_settings: pydantic_settings.PydanticBaseSettingsSource,
        env_settings: pydantic_settings.PydanticBaseSettingsSource,
        dotenv_settings: pydantic_settings.PydanticBaseSettingsSource,
        file_secret_settings: pydantic_settings.PydanticBaseSettingsSource,
    ) -> typing.Tuple[pydantic_settings.PydanticBaseSettingsSource, ...]:
        return (
            pydantic_settings.YamlConfigSettingsSource(settings_cls),
            pydantic_settings.EnvSettingsSource(settings_cls),
        )


@functools.lru_cache()
def config() -> Config:
    cfg = Config()
    return cfg



import asyncio
import datetime
import functools
import inspect
import random
import traceback
from typing import Optional

from .api_base import RetryError, RPMLimitError, UnrecoverableAPIError, log


def retry(max_attempts=3, initial_delay=1, backoff_factor=2):
    def decorator(func):
        if not inspect.iscoroutinefunction(func):
            raise ValueError(
                f"api_retry decorator only applies to async function, {func.__name__} is not an async function"
            )

        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            attempts = 1
            delay = initial_delay
            last_error = None
            stack_trace: Optional[str] = None

            while attempts <= max_attempts:
                call_time = datetime.datetime.now()
                try:
                    if inspect.iscoroutinefunction(func):
                        return await func(*args, **kwargs)
                    else:
                        return func(*args, **kwargs)
                except UnrecoverableAPIError as e:
                    raise RetryError(attempts, max_attempts, e, traceback.format_exc())
                except RPMLimitError as err:
                    last_error = err
                    stack_trace = traceback.format_exc()
                    log.debug(f"api_retry: Attempt {attempts} out of {max_attempts}: {err}")
                    wait_for_s = err.wait_for_s
                    if not wait_for_s:
                        now = datetime.datetime.now()
                        if call_time.minute != now.minute:
                            continue
                        next_minute = now.replace(second=0, microsecond=0) + datetime.timedelta(minutes=1)
                        wait_for_s = (next_minute - now).total_seconds() + (3 + random.random() * 5)
                    if attempts < max_attempts:
                        log.debug(f"Attempting again in {wait_for_s}s.")
                        await asyncio.sleep(wait_for_s)
                except Exception as err:
                    log.debug(f"api_retry: Attempt {attempts} out of {max_attempts}: {err}")
                    last_error = err
                    stack_trace = traceback.format_exc()
                    if attempts < max_attempts:
                        await asyncio.sleep(delay)
                finally:
                    delay *= backoff_factor
                    attempts += 1

            raise RetryError(attempts - 1, max_attempts, last_error, stack_trace)

        return wrapper

    return decorator



import abc
import datetime
import textwrap
import typing
from typing import Union, Optional, Any
import pydantic


class EvaluationResult(pydantic.BaseModel):
    score: Optional[float] = pydantic.Field(description=textwrap.dedent(
        """\
        Score of the evaluation. Can be any numerical value, though typically ranges from 0 to 1,
        where 1 represents the best possible score. 
        """
    ))
    # Whether the evaluation is considered to pass or fail
    pass_: Optional[bool] = None
    # Text output of the evaluation.
    # Usually used for discrete human-readable category evaluation or as a label for score value.
    text_output: Optional[str] = None
    # Arbitrary json-serializable metadata about evaluation.
    metadata: Optional[dict[str, Any]] = None
    # Human-readable explanation of the evaluation.
    explanation: Optional[str] = None
    # Key-value pair metadata
    tags: Optional[dict[str, str]] = None
    # Duration of the evaluation.
    # In case value is not set, @evaluator decorator and Evaluator classes will set this value automatically.
    evaluation_duration: Optional[datetime.timedelta] = None
    # Duration of the evaluation explanation.
    explanation_duration: Optional[datetime.time] = None


class Evaluator(abc.ABC):
    @abc.abstractmethod
    def evaluate(self, *args, **kwargs) -> Union[Optional[EvaluationResult], typing.Awaitable[Optional[EvaluationResult]]]:
        ...


class Foo(Evaluator):
    def evaluate(self, *args, **kwargs) -> EvaluationResult:
        pass

    ...